theory Index
  imports "../heap_monad/Memory_Heap"
begin

lemma index_mono:
  fixes a b a0 b0 :: nat
  assumes a: "a < a0" and b: "b < b0"
  shows "a * b0 + b < a0 * b0"
proof -
  have "a * b0 + b < (Suc a) * b0"
    using b by auto
  also have "\<dots> \<le> a0 * b0"
    using a[THEN Suc_leI, THEN mult_le_mono1] .
  finally show ?thesis .
qed

lemma index_eq_iff:
  fixes a b c d b0 :: nat
  assumes "b < b0" "d < b0" "a * b0 + b = c * b0 + d"
  shows "a = c \<and> b = d"
proof (rule conjI)
  { fix a b c d :: nat
    assume ac: "a < c" and b: "b < b0"
    have "a * b0 + b < (Suc a) * b0"
      using b by auto
    also have "\<dots> \<le> c * b0"
      using ac[THEN Suc_leI, THEN mult_le_mono1] .
    also have "\<dots> \<le> c * b0 + d"
      by auto
    finally have "a * b0 + b \<noteq> c * b0 + d"
      by auto
  } note ac = this

  { assume "a \<noteq> c"
    then consider (le) "a < c" | (ge) "a > c"
      by fastforce
    hence False proof cases
      case le show ?thesis using ac[OF le assms(1)] assms(3) ..
    next
      case ge show ?thesis using ac[OF ge assms(2)] assms(3)[symmetric] ..
    qed
  }
  
  then show "a = c"
    by auto

  with assms(3) show "b = d"
    by auto
qed

instantiation prod :: (order, order) order
begin

fun less_prod :: "'a \<times> 'b \<Rightarrow> 'a \<times> 'b \<Rightarrow> bool" where
  "less_prod (a,b) (c,d) = (a < c \<and> b < d)"
fun less_eq_prod :: "'a \<times> 'b \<Rightarrow> 'a \<times> 'b \<Rightarrow> bool" where
  "less_eq_prod ab cd = (ab < cd \<or> ab = cd)"

instance by standard fastforce+
end

datatype 'a bound = Bound (lower: 'a) (upper:'a)

definition in_bound :: "'a :: order bound \<Rightarrow> 'a \<Rightarrow> bool" where
  "in_bound bound x \<equiv> case bound of Bound l r \<Rightarrow> l \<le> x \<and> x < r"

class index = order +
  fixes idx :: "'a::order bound \<Rightarrow> 'a \<Rightarrow> nat" and size :: "'a bound \<Rightarrow> nat"
  assumes idx_valid: "in_bound bound x \<Longrightarrow> idx bound x < size bound"
    and idx_inj : "\<lbrakk>in_bound bound x; in_bound bound y; idx bound x = idx bound y\<rbrakk> \<Longrightarrow> x = y"

locale bounded_index =
  fixes bound :: "'k :: index bound"
begin

definition "size \<equiv> index_class.size bound" for size

definition "checked_idx x \<equiv> if in_bound bound x then idx bound x else size"

lemma checked_idx_injective:
  "injective size checked_idx"
  unfolding injective_def
  unfolding checked_idx_def
  using idx_inj by (fastforce split: if_splits)
end

instantiation prod :: (index, index) index begin
fun idx_prod :: "('a \<times> 'b) bound \<Rightarrow> 'a \<times> 'b \<Rightarrow> nat" where
  "idx_prod (Bound (l0, r0) (l1, r1)) (a, b) = (idx (Bound l0 l1) a) * (size (Bound r0 r1)) + idx (Bound r0 r1) b"

fun size_prod :: "('a \<times> 'b) bound \<Rightarrow> nat" where
  "size_prod (Bound (l0, r0) (l1, r1)) = size (Bound l0 l1) * size (Bound r0 r1)"

instance proof
  { fix ab :: "'a \<times> 'b" and bound :: "('a \<times> 'b) bound"
    assume bound: "in_bound bound ab"

    obtain a b l0 r0 l1 r1 where defs:"ab = (a, b)" "bound = Bound (l0, r0) (l1, r1)"
      by (cases ab; cases bound) auto

    with bound have a: "in_bound (Bound l0 l1) a" and b: "in_bound (Bound r0 r1) b"
      unfolding in_bound_def by auto

    have "idx (Bound (l0, r0) (l1, r1)) (a, b) < size (Bound (l0, r0) (l1, r1))"
      using index_mono[OF idx_valid[OF a] idx_valid[OF b]] by auto

    thus "idx bound ab < size bound"
      unfolding defs .
  }

  { fix ab cd :: "'a \<times> 'b" and bound :: "('a \<times> 'b) bound"
    assume bound: "in_bound bound ab" "in_bound bound cd"
      and idx_eq: "idx bound ab = idx bound cd"

    obtain a b c d l0 r0 l1 r1 where
      defs: "ab = (a, b)" "cd = (c, d)" "bound = Bound (l0, l1) (r0, r1)"
      by (cases ab; cases cd; cases bound) auto

    from defs bound have
          a: "in_bound (Bound l0 r0) a"
      and b: "in_bound (Bound l1 r1) b"
      and c: "in_bound (Bound l0 r0) c"
      and d: "in_bound (Bound l1 r1) d"
      unfolding in_bound_def by auto

    from index_eq_iff[OF idx_valid[OF b] idx_valid[OF d] idx_eq[unfolded defs, simplified]]
    have ac: "idx (Bound l0 r0) a = idx (Bound l0 r0) c" and bd: "idx (Bound l1 r1) b = idx (Bound l1 r1) d" by auto
    show "ab = cd"
      unfolding defs using idx_inj[OF a c ac] idx_inj[OF b d bd] by auto
  }
qed
end

instantiation nat :: index
begin
fun idx_nat :: "nat bound \<Rightarrow> nat \<Rightarrow> nat" where
  "idx_nat (Bound l _) i = i - l"

fun size_nat :: "nat bound \<Rightarrow> nat" where
  "size_nat (Bound l r) = r - l"

instance proof
qed (auto simp: in_bound_def split: bound.splits) 
end

instantiation int :: index
begin
fun idx_int :: "int bound \<Rightarrow> int \<Rightarrow> nat" where
  "idx_int (Bound l _) i = nat (i - l)"

fun size_int :: "int bound \<Rightarrow> nat" where
  "size_int (Bound l r) = nat (r - l)"

instance proof
qed (auto simp: in_bound_def split: bound.splits) 
end

paragraph \<open>Code Setup\<close>
lemmas [code] =
  bounded_index.checked_idx_def
  bounded_index.size_def

end